//modID=2193707 modFileID=2779841
//modPlatform=Modio

// East Asia map script for Old World with two Great Rivers
// Made by Harry Gao
// V2.1
//
// This map script borrows from Seaside, to create a map with a twin river basin 
// single player map
// todos: extra barb camp level depend on barb strength setting

using Mohawk.SystemCore;
using TenCrowns.GameCore;
using UnityEngine;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Globalization;

public class MapScriptEastAsia : DefaultMapScript
{
	// map params as supported by parent
	public override int DistanceFromLandToSetWaterAsBoundary { get; set; } = 5; // Set to a high number, such as 200, to prevent the setting of any water tiles as out of bounds.
	public override DirectionType PrimaryWindDirection { get; set; } = DirectionType.E;
	public override int PrimaryRainStrength { get; set; } = 4;

	public override int NumMaxAdjacentTundraAtCitySite { get; set; } = 100;
	public override int SpacingForDesertTags { get { return 15; } }
	public override int SpacingForIslandTags { get { return 4; } }
	public override int SpacingForRiverTags { get { return 10; } }
	public override int MinRiverLengthForNaming { get; set; } = 15;
	public override int MinLakeSizeForNaming { get; set; } = 20;
	public override int MinIslandSizeForNaming { get; set; } = 4;
	public override int MaxIslandSizeForNaming { get; set; } = 100;
	public override int MapBoundaryCornerRadius { get; set; } = 3;

	public override bool AllowMarshCitySite { get; set; } = true;
	public override bool AllowTundraCitySite { get; set; } = true;

	readonly MapOptionType TRIBES_RANDOM = MapOptionType.NONE;
	readonly MapOptionType TRIBES_SCATTERED = MapOptionType.NONE;
	readonly MapOptionType TRIBES_QUADRANTS = MapOptionType.NONE;
	readonly MapOptionType TRIBES_INLAND = MapOptionType.NONE;
	readonly MapOptionType TRIBES_COASTAL = MapOptionType.NONE;
	readonly MapOptionType TUNDRA = MapOptionType.NONE;
	readonly MapOptionType DESERT = MapOptionType.NONE;
	readonly MapOptionType MOUNTAINS = MapOptionType.NONE;
	readonly MapOptionType WATER = MapOptionType.NONE;
	readonly MapOptionType ONE = MapOptionType.NONE;
	readonly MapOptionType TWO = MapOptionType.NONE;
	readonly MapOptionType THREE = MapOptionType.NONE;
	readonly MapOptionType HORDE = MapOptionType.NONE;
	readonly MapOptionType WEAK = MapOptionType.NONE;

	readonly int HillsPercent = 13;
	readonly int RiverMinPercent = 7;
	private int numOfGreatRivers = 2;

	MapOptionType eTerrainOption = MapOptionType.NONE;
	MapOptionType eTribeOption = MapOptionType.NONE;
	MapOptionType eRiverCount = MapOptionType.NONE;
	MapOptionType eBarbarianMultiCamp = MapOptionType.NONE; //additional options require xml modifcation alongside scripts; waiting for more advanced modding capabilities

	HashSet<TileData> greatRiverTiles = new HashSet<TileData>();

	public static string GetName()
	{
		return "The Middle Kingdom";
	}

	public static string GetHelp()
	{
		return
			"A coastal basin cradled by large, overflowing rivers, " +
			"with a coastline to the East and South. A few fierce Northern Tribes rule the vast " +
			"wastelands, and barbarian hordes descent from the Himalayas from to the West in" +
			" great numbers. Will you be able to rule the known Old World, or will your armies " +
			"be stopped by the many natural obstacles stand in your way? Topology inspired by that of China.";
	}

	protected override void SetMapSize()
	{
		base.SetMapSize();
		// map needs to be bigger. Got a lot of unused border areas...probably.
		if (eTerrainOption == WATER || eTerrainOption == DESERT)
		{
			mapParameters.iWidth += ((mapParameters.iWidth * 30) / 100);
		}
		else
			mapParameters.iWidth += ((mapParameters.iWidth * 10) / 100);
		mapParameters.iHeight += ((mapParameters.iHeight * 20) / 100);


		numOfGreatRivers = (mapParameters.iHeight - 25) / 25;


	}


	public new static void GetCustomOptionsMulti(List<MapOptionsMultiType> options, Infos infos)
	{
		options.Add(infos.getType<MapOptionsMultiType>("MAP_OPTIONS_MULTI_SEASIDE_TERRAIN_TYPE_VARIATION"));
		options.Add(infos.getType<MapOptionsMultiType>("MAP_OPTIONS_MULTI_CONTINENT_TRIBES"));

		DefaultMapScript.GetCustomOptionsMulti(options, infos);
	}

	public MapScriptEastAsia(ref MapParameters mapParameters, Infos infos) : base(ref mapParameters, infos)
	{
		TUNDRA = infos.getType<MapOptionType>("MAP_OPTION_SEASIDE_TERRAIN_TUNDRA");
		DESERT = infos.getType<MapOptionType>("MAP_OPTION_SEASIDE_TERRAIN_DESERT");
		MOUNTAINS = infos.getType<MapOptionType>("MAP_OPTION_SEASIDE_TERRAIN_MOUNTAINS");
		WATER = infos.getType<MapOptionType>("MAP_OPTION_SEASIDE_TERRAIN_WATER");
		TRIBES_RANDOM = infos.getType<MapOptionType>("MAP_OPTION_CONTINENT_TRIBES_RANDOM");
		TRIBES_SCATTERED = infos.getType<MapOptionType>("MAP_OPTION_CONTINENT_TRIBES_SCATTERED");
		TRIBES_QUADRANTS = infos.getType<MapOptionType>("MAP_OPTION_CONTINENT_TRIBES_QUADRANTS");
		TRIBES_INLAND = infos.getType<MapOptionType>("MAP_OPTION_CONTINENT_TRIBES_INLAND");
		TRIBES_COASTAL = infos.getType<MapOptionType>("MAP_OPTION_CONTINENT_TRIBES_COASTAL");
	}

	protected override void InitMapData()
	{
		base.InitMapData();

		MapOptionsMultiType eTerrain = infos.getType<MapOptionsMultiType>("MAP_OPTIONS_MULTI_SEASIDE_TERRAIN_TYPE_VARIATION");
		if (!mapParameters.mapMapMultiOptions.TryGetValue(eTerrain, out eTerrainOption))
		{
			eTerrainOption = MapOptionType.NONE;
		}
		MapOptionsMultiType eTribe = infos.getType<MapOptionsMultiType>("MAP_OPTIONS_MULTI_CONTINENT_TRIBES");
		if (!mapParameters.mapMapMultiOptions.TryGetValue(eTribe, out eTribeOption))
		{
			eTribeOption = MapOptionType.NONE;
		}
		MinLatitude = (short)(5 + random.Next(21)); // Range from 5-25
		MaxLatitude = (short)(40 + random.Next(31)); // Range from 40 to 70

		if (eTerrainOption == TUNDRA)
		{
			MinLatitude = (short)(13 + MinLatitude / 2);
			MaxLatitude = (short)(35 + MaxLatitude / 2);
			PrimaryRainStrength += 1;
			RainShadowAdjustment -= 1;
		}
		else if (eTerrainOption == DESERT)
		{
			MinLatitude /= 2;
			MaxLatitude = (short)(20 + MaxLatitude / 2); //45-60
			PrimaryRainStrength -= 1;
			RainShadowAdjustment += 3;
		}

		// Water depends on terrain option:  Extra Water : Anything else.
		OceanPercent = (short)(16 + random.Next(10));
		LakePercent = (short)random.Next(4);

		if (eTerrainOption == WATER)
		{
			OceanPercent += (short)random.Next(4);
			LakePercent += (short)random.Next(8);
			PrimaryRainStrength += 1;
			RainShadowAdjustment -= 2;
		}

		CoastPercent = (short)(OceanPercent / 2);

	}
	protected override void GenerateLand()
	{
		using (new UnityProfileScope("DefaultMapScript.SetLand"))
		{
			heightGen = MakeTopology();
			heightGen.Normalize();

			// Set the lowest tiles to be water
			List<NoiseGenerator.TileValue> tiles = heightGen.GetPercentileRange(0, OceanPercent);

			foreach (NoiseGenerator.TileValue tile in tiles)
			{
				TileData loopTile = GetTile(tile.x, tile.y);

				loopTile.meTerrain = WATER_TERRAIN;
				loopTile.meHeight = infos.Globals.OCEAN_HEIGHT;
			}

			// Set the next lowest tiles to be coast
			heightGen.AddNoise(MapWidth / (1 + random.Next(11)), random.Next(100) / 600f);
			heightGen.Normalize();
			tiles = heightGen.GetPercentileRange(OceanPercent, OceanPercent + CoastPercent * 2);

			foreach (NoiseGenerator.TileValue tile in tiles)
			{
				if (tile.value > random.Next(100) / 1000f + (eTerrainOption == WATER ? 0.18 : 0.09)) //too tall
					break;
				TileData loopTile = GetTile(tile.x, tile.y);
				//near land? should reduce holes and create peninsulae 
				bool skipWater = false;
				if (tile.y > MapHeight / 2 || tile.x < MapWidth / 4 * 3)
				{
					TileData nearbyTile = random.Next(3) > 0 ? TileAdjacent(loopTile, DirectionType.SW) :
						random.Next(tile.x / 20) > 1 ? TileAdjacent(loopTile, DirectionType.NE) :
						GetTile(tile.x - 4, tile.y);
					TileData west = GetTile(tile.x - 1, tile.y);
					skipWater = nearbyTile != null && nearbyTile.meTerrain != WATER_TERRAIN && west != null && west.meTerrain != WATER_TERRAIN;

				}
				if (!skipWater)
				{
					loopTile.meTerrain = WATER_TERRAIN;
					loopTile.meHeight = infos.Globals.COAST_HEIGHT;
				}
			}

			// Hills
			// Scattered hills, not part of mountain chains.
			heightGen.AddNoise(MapWidth / (1 + random.Next(4) + random.Next(4)), eTerrainOption == MOUNTAINS ? 0.8f : 0.4f);
			heightGen.AddNoise(MapWidth / (5 + random.Next(12) + random.Next(12)), 0.7f);

			heightGen.Normalize();

			tiles = heightGen.GetPercentileRange(100 - HillsPercent, 100);
			foreach (NoiseGenerator.TileValue tile in tiles)
			{
				TileData loopTile = GetTile(tile.x, tile.y);

				if (loopTile.meHeight == infos.Globals.DEFAULT_HEIGHT)
				{

					if (tile.value > (eTerrainOption == MOUNTAINS ? 0.8 : 0.9) + random.Next(7) / 100f)
					{
						loopTile.meHeight = infos.Globals.MOUNTAIN_HEIGHT;
					}
					else
						loopTile.meHeight = infos.Globals.HILL_HEIGHT;
				}
			}

			//mark outer boundaries
			foreach (var tile in Tiles)
			{
				if (TileX(tile) < 2 || TileX(tile) > MapWidth - 3 || TileY(tile) < 2 || TileY(tile) > MapHeight - 3)
				{
					tile.mbBoundary = true;
				}
			}
			ResetDistances();
		}
	}

	/**
	 * Create the overall topological shape--resembling East Asia
	 * 
	 */
	private NoiseGenerator MakeTopology()
	{
		NoiseGenerator hg = new NoiseGenerator(MapWidth, MapHeight, random.NextSeed());

		for (int x = 0; x < MapWidth; ++x)
		{
			for (int y = 0; y < MapHeight; ++y)
			{
				int i3D6 = 3 + random.Next(6) + random.Next(6) + random.Next(6); //this is 3D6
				int i2D4 = 2 + random.Next(4) + random.Next(4); //this is 2D4 
				int i2D2 = 2 + random.Next(2) + random.Next(2); //2-4
				double curve = (Math.Sin(x * y * 5 * Math.PI / MapWidth / MapHeight) + 1) / 2; //0 to 1

				//base topology
				float h = Math.Abs(MapHeight - x - y / 3);
				h += (x * 1.4f + Math.Abs(MapHeight / 3 - y) * curve + i2D4 < MapWidth / 2 ? i3D6 * x / 2 : 0); //West
				h /= (x + i3D6 - y * curve > MapWidth / 2.5f) ? i2D4 : 1; //SE flattening

				//blurring oceans
				h = DifferentiatingWater(y / 4, i2D4, h, 35);

				//making taiwan
				if (x + i2D2 > MapWidth / 1.5f && y + i2D4 < MapHeight / 4)
				{
					float taiwanModifier = (float)Math.Sqrt(Math.Pow(x - MapWidth / 1.5f, 2) + Math.Pow(y - MapHeight / 4f, 2) / 2);
					//	Console.WriteLine(taiwanModifier);
					if (taiwanModifier < MapWidth / 5.8 && taiwanModifier > MapWidth / (9 - i2D2))
						h += taiwanModifier * i3D6;
				}

				//mainland
				float NEModifier = (float)(MapWidth - Math.Sqrt(Math.Pow(x - MapWidth / 2f, 2) + Math.Pow(y - MapHeight / 1.8f, 2) / (3 + curve)));
				if (NEModifier + i2D2 > 0.8f * MapWidth)
				{
					h += NEModifier / 2;
				}
				if (y > MapHeight / 2 && x - i3D6 < MapWidth / 2)
				{
					h += (int)Math.Max(0, MapWidth - (x - y / 2) * curve);
				}

				//making shandong
				if (x + i2D4 > MapWidth / 2 && y - i2D4 > MapHeight / 2)
				{
					h += (float)Math.Max(-35,
						 MapHeight - Math.Pow(y - 0.33f * MapHeight, 2) / 2);

					h *= h < 0 ? -0.5f : 1;
				}

				//making Korea			
				if (x - y / 10 + i3D6 + i2D2 > MapWidth && y + i2D4 > MapHeight / 2)
				{
					NEModifier = (float)Math.Sqrt(
						Math.Pow(x - MapWidth * 0.6, 2) + Math.Pow(y - MapHeight * 0.6, 2));

					if (NEModifier > MapWidth / (10 - i2D2))
						h += NEModifier * i2D4;
				}
				//blurring oceans
				h = DifferentiatingWater(Math.Max(2, y / 3 + i3D6 - MapHeight / 5), i2D4, h, x / 3 + y / 2);

				if (y + i2D4 < MapHeight / 5 || y + i3D6 > MapHeight)
					h *= 1.2f;
				if ((x - i2D4) * 3 / MapWidth == 1)
				{
					h /= 2f;
				}

				h /= MapWidth / 7; //normalize
				h = Math.Min(8, h); //truncate extremes


				hg.SetValue(x, y, h);
			}

		}
		return hg;
	}

	private static float DifferentiatingWater(int heightBoost, int noise, float height, int coastLine)
	{
		if (height < coastLine + noise)
			height /= 3;//if you low, you get REALLY low. 
		height += Math.Max(0, heightBoost - noise);
		if (height < coastLine + noise)
			height /= 3;
		return height;
	}
	protected override void GenerateMountains()
	{
		//	Debug.Log("genearting mountains");
		using (new UnityProfileScope("DefaultMapScript.GenerateMountains"))
		{
			// Amount of mountains depends on terrain option.

			dTargetPercentage = (short)random.Next(4) + (short)random.Next(4) + eTerrainOption == MOUNTAINS ? 5 : 0;

			dDisobedienceThreshold = eTerrainOption == MOUNTAINS ? (short)(63) : (short)(84);
			iGapMin = eTerrainOption == MOUNTAINS ? (short)(10) : (short)(5);
			iGapMax = eTerrainOption == MOUNTAINS ? (short)(20) : (short)(15);

			iMaxNumberOfTilesInChain = MapHeight / 6;

			base.GenerateMountains();

			foreach (TileData tile in Tiles)
			{

				if (tile.meHeight < infos.Globals.VOLCANO_HEIGHT &&
					(CountAdjacent(tile, IsMountainAny) > 5 || CountAdjacent(tile, IsMountainOrHillsAny) > 3 + random.Next(4)))
				{
					tile.meHeight = infos.Globals.MOUNTAIN_HEIGHT;
					tile.meTerrain = tile.meTerrain == WATER_TERRAIN ? TEMPERATE_TERRAIN : tile.meTerrain;
				}
			}
			ResetDistances();
		}
	}

	protected override void GenerateDeserts()
	{
		base.GenerateDeserts();
		int extreme = (MapHeight / 4);
		int borderline = (MapHeight * 4 / 5);
		int temperate = (int)(extreme / 4);
		int xmin = MapWidth / 4;
		int buffer = 3;
		int xmax = MapWidth - buffer;

		if (eTerrainOption == DESERT)
			xmax -= MapWidth / 3;
		else if (eTerrainOption == TUNDRA)
			extreme = (int)(extreme * 2);
		else eTerrainOption = TUNDRA; //make tundra if neither option is picked

		heightGen.AddNoise(10f, 0.6f);

		for (int x = xmin - buffer; x < xmax + buffer; x++)
		{
			int extreme_range_for_this_x = (int)(extreme * heightGen.GetValue(x, extreme)) - temperate;
			int borderline_range_for_this_x = (int)(borderline * heightGen.GetValue(x, borderline)) - temperate;
			extreme_range_for_this_x = Math.Max(0, extreme_range_for_this_x);
			borderline_range_for_this_x = Math.Max(extreme_range_for_this_x, borderline_range_for_this_x);

			for (int y = MapHeight - borderline_range_for_this_x; y < MapHeight; ++y)
			{
				TileData tile = GetTile(x, y);
				if (tile == null) break;
				if (!IsWater(tile))
				{
					tile.meTerrain = TROPICAL_TERRAIN;
				}
			}
			for (int y = MapHeight - extreme_range_for_this_x; y < MapHeight; ++y)
			{
				TileData tile = GetTile(x, y);
				if (tile == null) break;
				if (x < xmin + random.Next(8) || x > xmax - random.Next(8))
					y += 3;
				else if (!IsWater(tile))
				{
					tile.meTerrain = eTerrainOption == TUNDRA ? TUNDRA_TERRAIN : DESERT_TERRAIN;
				}
			}
		}
		ShrinkDesert();
	}
	protected override void GenerateRivers()
	{
		using (new UnityProfileScope("DefaultMapScript.GenerateRivers"))
		{
			DoRiverConnections();
			List<TileData> riverSourceCandidates = GetRiverSources().ToList();

			//normal river making	
			int riverTileCount = 0;
			int minGRLength = MapWidth / 3;
			var greatRiverCandidates = new SortedDictionary<int, List<TileData>>();
			while ((greatRiverCandidates.Count < 2 || riverTileCount < RiverMinPercent / 100f * Tiles.Count) && riverSourceCandidates.Count > 0)
			{
				CalculateDistancesFromRiver();
				int bestValue = int.MinValue;
				TileData bestTile = null;
				foreach (TileData tile in riverSourceCandidates)
				{
					int distanceFromRiverOrWater = GetDistanceFromRiver(tile);
					if (distanceFromRiverOrWater > bestValue)
					{
						bestValue = distanceFromRiverOrWater;
						bestTile = tile;
					}
				}
				riverSourceCandidates.Remove(bestTile);

				List<TileData> riverTiles = new List<TileData>();
				foreach (TileData riverTile in StartRiver(bestTile, true))
				{
					riverTiles.Add(riverTile);
				}

				if (riverTiles.Count > minGRLength) //Actual river length is about half the count since both banks are included in the count
				{
					riverTileCount -= greatRiverCandidates.Count < numOfGreatRivers * 2 ? riverTiles.Count / numOfGreatRivers / 2 : 0; //great rivers are freebies that don'seed count toward total river percentage; rough math
					int score = riverTiles.Count * 20; //longer the better
					score -= TileX(riverTiles[0]) * 8; //start West
					TileData end = riverTiles[riverTiles.Count - 1];
					score += 8 * base.TileX(end); //end East								 
					score -= 15 * (Math.Abs(TileY(end) - TileY(riverTiles[0]))); //prefer rivers end and start at the same height
					score -= 3 * (int)Math.Pow(base.MapHeight / 2 - base.TileY(riverTiles[random.Next(minGRLength)]), 2) / numOfGreatRivers; //not too far North or South; can be more forgiving if there are a lot of rivers

					while (!greatRiverCandidates.TryAdd(score, riverTiles))
						score--;
				}
				riverTileCount += riverTiles.Count;

				terrainAreas.Add(new TerrainArea(AreaTerrainType.RIVER, riverTiles.Select(t => (int)t.ID).ToList(), this));
			}
			if (greatRiverCandidates.Count >= numOfGreatRivers)
			{
				var key = greatRiverCandidates.Keys.Last();
				var value = greatRiverCandidates[key];
				greatRiverCandidates.Remove(key);
				MakeRiverGreat(value); //first river

				for (int i = 1; i < numOfGreatRivers; i++)
				{
					if (greatRiverCandidates.Count < 1)
						break;
					var greatRiverCandidates2 = new SortedDictionary<int, List<TileData>>();
					foreach (var candidate in greatRiverCandidates)
					{
						int sample = random.Next(minGRLength);
						int newScore = candidate.Key;
						newScore += 5 * (Math.Abs(base.TileY(value[sample]) - base.TileY(candidate.Value[sample])));
						newScore -= 59 * (greatRiverTiles.Intersect(candidate.Value).Count()); //no intersections please.
						while (!greatRiverCandidates2.TryAdd(newScore, candidate.Value))
							newScore--;
					}
					MakeRiverGreat(greatRiverCandidates2.Values.Last()); //more rivers
					greatRiverCandidates2.Remove(greatRiverCandidates2.Keys.Last());
					greatRiverCandidates = greatRiverCandidates2;
				}
			}
			else
				Debug.Log("Failed to find any suitable great river candidates");
		}
	}

	private void MakeRiverGreat(List<TileData> riverTiles)
	{

		riverTiles = ExtendRiver(false, riverTiles);
		FixCoast();
		riverTiles = ExtendRiver(true, riverTiles);
		greatRiverTiles.AddRange(riverTiles);
		int clustering = 0;
		bool firsts = true;
		for (int i = 0; i < riverTiles.Count; i++)
		{
			//	Debug.Log(String.Format("rivertiles at {0}: ({1}, {2})", i, TileX(riverTiles[i]), TileY(riverTiles[i])));
			if (i + 10 > riverTiles.Count)
				clustering++;
			TileData riverTile = riverTiles[i];
			// river thicker to the East, but a small chance of staying thin 
			if (firsts)
			{
				firsts = random.Next(4) == 0;
				if (GetDistanceFromBoundary(riverTile) > 1)
				{
					PlaceLandInRiver(TileAdjacent(riverTile, DirectionType.SW), true);
					PlaceLandInRiver(TileAdjacent(riverTile, DirectionType.W), true);
					PlaceLandInRiver(TileAdjacent(riverTile, DirectionType.NW), true);
				}
				PlaceLandInRiver(riverTile, true);
			}
			else if (Math.Sqrt(i) - CountLake(riverTile) / 2 * 4 >
				(clustering * random.Next(4) + random.Next(3)) / 2 * random.Next(20) + CountRiverEdges(riverTile))
			{
				//make it a "lake" 
				riverTile.meTerrain = WATER_TERRAIN;
				riverTile.meHeight = infos.Globals.LAKE_HEIGHT;
				clustering = 0;
			}
			else
			{
				PlaceLandInRiver(riverTile, false);
				clustering++;
			}
		}
		ConnectRiverParts(riverTiles);
	}

	private List<TileData> ExtendRiver(bool forward, List<TileData> riverTiles)
	{
		if (forward)
		{
			TileData tile = riverTiles.Last();
			if (GetDistanceFromSaltWater(tile) == 0)
			{ //if the river naturally ends at the coast
				return riverTiles;
			}
			while (GetDistanceFromBoundary(tile) > 1)
			{
				int dir = random.Next(2) + random.Next(2); //SE, E and NE
				if (TileY(tile) > MapHeight / 4)
					dir--;
				else if (TileY(tile) < MapHeight / 4)
					dir++;
				switch (dir)
				{
					case -1:
					case 0: //SE				
						tile = TileAdjacent(tile, DirectionType.SE);
						break;
					case 1: //E
						tile = TileAdjacent(tile, DirectionType.E);
						break;
					case 2: //NE
					case 3:
						tile = TileAdjacent(tile, DirectionType.NE);
						break;
					default:
						break;
				}
				riverTiles.Add(tile);
				if (CountAdjacent(tile, IsCoast) > 0)
				{
					tile.meHeight = infos.Globals.COAST_HEIGHT;
					if (!ConvertArea(tile, infos.Globals.LAKE_HEIGHT, 40))
					{
						//a lot of coastal tiles discovered. this is not a river anymore
						break;
					}
				}
				if (CountAdjacent(tile, IsOcean) > 0)
				{
					tile.meHeight = infos.Globals.OCEAN_HEIGHT;
					if (!ConvertArea(tile, infos.Globals.LAKE_HEIGHT, 10))
					{
						break;
					}
				}

				var widener = TileAdjacent(tile, (DirectionType)(dir + 1 + random.Next(2)));
				if (widener != null && random.Next(2) == 0)
				{
					riverTiles.Add(widener);
					if (CountAdjacent(widener, IsCoast) > 0)
					{
						widener.meHeight = infos.Globals.COAST_HEIGHT;
						if (!ConvertArea(widener, infos.Globals.LAKE_HEIGHT, 40))
						{
							//a lot of coastal tiles discovered. this is not a river anymore
							break;
						}
					}
					if (CountAdjacent(widener, IsOcean) > 0)
					{
						widener.meHeight = infos.Globals.OCEAN_HEIGHT;
						if (!ConvertArea(widener, infos.Globals.LAKE_HEIGHT, 10))
						{
							break;
						}
					}
				}
			}
			return riverTiles;
		}

		else //tracing from start of the river back to a mountain
		{
			List<TileData> precursors = new List<TileData>();
			TileData tile = riverTiles.First();
			int i;
			for (i = 0; i < MapWidth / 3; i++)
			{
				if (tile == null || CountAdjacent(tile, IsMountainAny) > 0 || GetDistanceFromBoundary(tile) < 1)
					break;

				//create a not so random walk toward center of Y
				precursors.Add(tile);
				int centerGravity = (TileY(tile) - MapHeight / 2); //big magnitude means farther from center, positve is North
				int bias = (random.Next(3) + random.Next(3) + random.Next(2) + random.Next(2)) - 3; //-3 to 3
				bias /= Math.Abs(centerGravity) > (MapHeight / 4) ? 2 : 1; //if gravity is strong, bias is small
				var dir = (int)(DirectionType.W - Math.Sign(centerGravity) + bias + (int)DirectionType.NUM_TYPES) % (int)DirectionType.NUM_TYPES;
				tile = TileAdjacent(tile, (DirectionType)dir);
			}

			precursors.Reverse(); //do we need to trim the last element? it's a duplicate of first element of rivertiles
			precursors.AddRange(riverTiles);
			return precursors;
		}
	}

	private void PlaceLandInRiver(TileData riverTile, bool tall)
	{

		if (riverTile == null)
			return;

		if (tall)
			riverTile.meHeight = infos.Globals.MOUNTAIN_HEIGHT;
		else
		{
			if (riverTile.meHeight != infos.Globals.LAKE_HEIGHT) //we refuse to make lakes into land...honoring other great rivers and RNGesus
				riverTile.meHeight = infos.Globals.VOLCANO_HEIGHT;
		}
	}

	private int CountLake(TileData tile)
	{
		int count = 0;
		for (DirectionType dir = 0; dir < DirectionType.NUM_TYPES; dir++)
		{
			TileData adjacent = TileAdjacent(tile, dir);
			if (adjacent != null && adjacent.meHeight == infos.Globals.LAKE_HEIGHT)
			{
				count++;
			}
		}
		return count;
	}

	protected override void GenerateElevations()
	{
		ResetDistances();
		List<List<int>> areas = new List<List<int>>();
		BuildAreas(areas, null, x => !x.mbBoundary && !x.isImpassable(infos));
		int tolerance = 0;
		while (areas[0].Count * 1.0 / Tiles.Count < 0.8 - tolerance / 100.0)
		{
			//need more tiles

			for (int i = 1; i < areas.Count; i++)
			{
				bool drilled = false;
				foreach (int tileID in areas[i])
				{
					TileData t = GetTile(tileID);
					TileData eastishNeighbor = TileAdjacent(t, DirectionType.NE + random.Next(3));
					while (eastishNeighbor != null && !eastishNeighbor.mbBoundary && (eastishNeighbor.meHeight == infos.Globals.MOUNTAIN_HEIGHT || eastishNeighbor.meHeight == infos.Globals.HILL_HEIGHT))
					{
						eastishNeighbor.meHeight--;
						eastishNeighbor = TileAdjacent(t, DirectionType.NE + random.Next(2) + random.Next(3));
						drilled = true;

					}
					if (drilled && random.Next(2) == 0)
						break;
				}
			}
			tolerance++;
			areas.Clear();
			BuildAreas(areas, null, x => !x.mbBoundary && !x.isImpassable(infos));
		}

		EliminatePondJags(3);
		EliminatePondJags(4);
		EliminatePondJags(5);
		EliminateSingletonMountains();
		BuffMountains(); //may close off some of the areas just opened
		base.GenerateElevations();
	}

	protected override void ModifyTerrain()
	{
		foreach (TileData t in Tiles)
		{
			if (t.meHeight > infos.Globals.HILL_HEIGHT && !t.mbBoundary)
			{
				//this should reduce mountain deserts--more accurate terrain
				DumpRain(t, (DirectionType)random.Next(6), random.Next(3), true, true);
				if (CountAdjacent(t, IsMountainAny) == 4)
				{
					//if a mountain is next to exactly 2 nonmountains, it may be lowered, depends on the geometry 
					DirectionType flat = DirectionType.NONE;
					for (DirectionType i = 0; i < DirectionType.NUM_TYPES; i++)
					{
						var adjT = TileAdjacent(t, i);
						if (adjT != null && adjT.meHeight < infos.Globals.MOUNTAIN_HEIGHT)
						{
							if (flat == DirectionType.NONE)
								flat = i;
							else switch (i - flat)
								{
									case 1:
									case 5:
										break;
									case 2:
									case 4:
										if (random.Next(2) == 0)
											break;
										else goto case 3;
									case 3:

										t.meHeight -= 2;
										adjT.mbRoad = adjT.meTerrain != WATER_TERRAIN;
										t.mbRoad = adjT.meTerrain != WATER_TERRAIN;
										TileAdjacent(t, flat).mbRoad = adjT.meTerrain != WATER_TERRAIN;
										break;
								}
						}
					}

				}

			}
			if (t.meTerrain == DESERT_TERRAIN && TileY(t) < MapHeight / 2)
			{
				//reduce Southern desert
				DumpRain(t, (DirectionType)random.Next(6), random.Next(2), true, true);
			}

			if (t.meHeight == infos.Globals.DEFAULT_HEIGHT && !t.mbBoundary)
			{
				if (CountAdjacent(t, IsSaltWater) + CountAdjacent(TileAdjacent(t, DirectionType.SW), IsSaltWater) > 5 + random.Next(3))
				{
					t.meHeight = infos.Globals.COAST_HEIGHT;
					t.meTerrain = WATER_TERRAIN;
				}
			}
			if (t.meHeight == infos.Globals.COAST_HEIGHT)
			{
				if (CountLake(t) > 0 && CountAdjacent(t, IsOcean) == 0)
				{
					ConvertArea(t, infos.Globals.LAKE_HEIGHT, 5);
				}
			}
		}
		foreach (TileData t in greatRiverTiles)
		{
			DumpRain(t, (DirectionType)random.Next(6), random.Next(3), true, true);
			if (t.meHeight == infos.Globals.VOLCANO_HEIGHT)
			{
				t.meHeight = infos.Globals.DEFAULT_HEIGHT;
				t.meTerrain = TEMPERATE_TERRAIN;
			}
			if (t.meTerrain == WET_TERRAIN && random.Next(2) == 0)
			{
				t.meTerrain = TROPICAL_TERRAIN;
			}
			DirectionType dir = (DirectionType)random.Next(6);

			for (int i = 0; i + dir < DirectionType.NUM_TYPES; i++)
			{
				var tile = TileAdjacent(t, i + dir);
				if (tile != null)
					DumpRain(tile, (DirectionType)random.Next(6), 6 - (int)dir, true, true);
			}
		}
	}

	private void EliminatePondJags(int threshhold = 4)
	{
		foreach (TileData tile in Tiles)
		{
			if (IsWater(tile) && CountAdjacent(tile, IsLand) > threshhold && !greatRiverTiles.Contains(tile))
			{
				tile.meHeight = infos.Globals.HILL_HEIGHT;
				tile.meTerrain = TEMPERATE_TERRAIN;
			}
		}
	}
	private void ConnectRiverParts(List<TileData> riverTiles)
	{
		TileData waterBody = null;

		for (int i = 0; i < riverTiles.Count; i++)
		{
			//		int x = TileX(riverTiles[i]);
			//		int y = TileY(riverTiles[i]);
			//		Debug.Log(String.Format("processing ({0}, {1})", x, y));
			if (IsLand(riverTiles[i]))
			{
				if (waterBody == null)
				{
					if (i > 0)
					{
						waterBody = riverTiles[i - 1];
					}
					else
					{
						waterBody = riverTiles[0];
					}
					//				Debug.Log(String.Format("set first water to be ({0}, {1})", TileX(waterBody), TileY(waterBody)));

				}
			}
			else if (waterBody != null && waterBody != riverTiles[i])
			{
				MakeRiverConnection(waterBody, riverTiles[i]);
				//		Debug.Log(String.Format("connected ({0}, {1}) to ({2}, {3})", TileX(waterBody), TileY(waterBody), x, y));
				waterBody = null;
			}
		}
		if (waterBody != null)
		{
			MakeRiverConnection(waterBody, riverTiles.Last());
			//	Debug.Log(String.Format("finally, connecting ({0}, {1}) to ({2}, {3})", TileX(waterBody), TileY(waterBody), TileX(riverTiles.Last()), TileY(riverTiles.Last())));
		}
	}

	//connect two tiles with rivers across any distance, designed for inputting two bodies of water,
	//althought the code adds some robustness around connecting land tiles too. The algorithm, simply 
	//put, is to connect the N corner of each hex en route from source to destination. River direction
	//is left to right. 
	private void MakeRiverConnection(TileData source, TileData destination, bool oppositeDir = false)
	{
		int distance = TileDistance(source.ID, destination.ID);
		if (distance < 1 || source.mbBoundary || destination.mbBoundary)
			return;

		int deltaX = TileX(destination) - TileX(source);

		var left = deltaX > 0 ? source : deltaX == 0 && !YIsEven(TileY(source)) ? source : destination;
		var right = deltaX > 0 ? destination : deltaX == 0 && !YIsEven(TileY(source)) ? destination : source;
		RotationType rot = deltaX > 0 ? RotationType.CLOCKWISE : deltaX == 0 && !YIsEven(TileY(source)) ? RotationType.CLOCKWISE : RotationType.COUNTER_CLOCKWISE;
		if (oppositeDir)
			rot = rot == RotationType.CLOCKWISE ? RotationType.COUNTER_CLOCKWISE : RotationType.CLOCKWISE;
		int deltaY = TileY(right) - TileY(left);
		var dir = deltaY == 0 ? DirectionType.E : deltaY > 0 ? DirectionType.NE : DirectionType.SE;

		TileData connector = TileAdjacent(left, dir);
		switch (dir)
		{
			case DirectionType.E:
				SetRiverRotation(left, DirectionType.NE, rot);
				SetRiverRotation(connector, DirectionType.NW, rot);
				break;
			case DirectionType.SE:
				SetRiverRotation(left, DirectionType.NE, rot);
				SetRiverRotation(left, DirectionType.E, rot);
				break;
			case DirectionType.NE:
				SetRiverRotation(connector, DirectionType.W, rot);
				SetRiverRotation(connector, DirectionType.NW, rot);
				break;
		}

		MakeRiverConnection(connector, right, right != destination);
	}


	//update a contiguous height to a different height 
	private bool ConvertArea(TileData seed, HeightType target, int max = int.MaxValue)
	{
		Stack<TileData> area = new Stack<TileData>();
		HeightType original = seed.meHeight;
		if (original == target)
		{
			return false;
		}
		area.Push(seed);
		while (area.Count > 0 && max > 0)
		{
			TileData nextTile = area.Pop();
			nextTile.meHeight = target;
			for (DirectionType dir = 0; dir < DirectionType.NUM_TYPES; dir++)
			{
				TileData adjacent = TileAdjacent(nextTile, dir);
				if (adjacent != null && adjacent.meHeight == original)
				{
					area.Push(adjacent);
				}
			}
			max--;
		}
		return max > 0;
	}

	private void RemoveRiverFromTile(TileData t, int num = 6)
	{
		for (DirectionType eLoopDirection = 0; eLoopDirection < DirectionType.NUM_TYPES; eLoopDirection++)
		{
			TileData pAdjacentTile = TileAdjacent(t, eLoopDirection);
			//remove extra river pieces
			if (pAdjacentTile != null && num > 0)
			{
				SetRiverRotation(t, eLoopDirection, RotationType.NONE);
				num--;
			}
		}
	}

	protected override void BuildWaterAreas()
	{
		infos.Globals.FRESH_WATER_THRESHOLD *= 3;//big rivers are still fresh water rivers
		base.BuildWaterAreas();
	}
	protected override bool IsBoundaryDeepTerrain(TileData tile)
	{
		if (tile == null)
			return true;
		if (tile.isImpassable(infos))
			return true;
		if (tile.meHeight > infos.Globals.HILL_HEIGHT)
			return true;
		if (tile.meTerrain == DESERT_TERRAIN || tile.meTerrain == TUNDRA_TERRAIN)
			return true; // random.Next(100) > 1;//1% chance of bad terrain being not bad enough to be marked out of bound
		return false;
	}
	protected override void PlaceTribes()
	{
		if (eTribeOption == TRIBES_SCATTERED) // Five Tribes, Dominance Method
		{
			PlaceFiveTribesUsingScatterMethod(3, true);
		}
		else if (eTribeOption == TRIBES_QUADRANTS) // Four Tribes, Quadrants method
		{
			PlaceFourTribesUsingQuadrantMethod(3, true, true);
		}
		else if (eTribeOption == TRIBES_INLAND) // Two Tribes, Inland method
		{
			PlaceTwoTribesUsingInlandMethod();
		}
		else if (eTribeOption == TRIBES_COASTAL) // Two Tribes, Coast method
		{
			tribesToUse.Clear();
			tribesToUse.Add(TRIBE_NUMIDIANS);
			tribesToUse.Add(TRIBE_SCYTHIANS);

			PlaceTwoTribesUsingCoastMethod(false);
		}
		else //TRIBES_RANDOM 
		{
			PlaceTribesUsingRandomlySelectedMethod();
		}

		AddExtraBarbarianCamps();
	}

	private void AddExtraBarbarianCamps()
	{
		//extra barbarians; must move player start to be able from these!
		int extraCampCount = 0;
		for (int y = 3; y < MapHeight - extraCampCount; y++)
		{ //Mountain barbs
			for (int x = 3; x * y < MapWidth * MapHeight / 4; x += (int)Math.Sqrt(y) + random.Next(1 + extraCampCount * 5))
			{
				TileData tile = GetTile(x, y);
				if (tile == null)
					break;
				if (!tile.isImpassable(infos) && !tile.mbBoundary)
				{
					if (tile.isUrban(infos) && tile.meImprovement != infos.Globals.RUINS_IMPROVEMENT)
					{
						PlaceBarbarianCamp(tile);
						extraCampCount++;
					}
					else if (0 == random.Next((y + x) / 3) && !tile.isWater(infos))
					{
						PlaceBarbarianCamp(tile);
						extraCampCount++;
					}
				}
			}
		}
		for (int x = MapWidth / 4 * 3; x < MapWidth - 2; x++)
		{ //Korean raiders
			for (int i = 3; i < MapHeight - 10; i += 2 * extraCampCount)
			{
				TileData tile = GetTile(x, i);
				if (tile != null && !tile.mbBoundary)
				{
					if (0 == random.Next(5) && (tile.meHeight == infos.Globals.DEFAULT_HEIGHT || tile.meHeight == infos.Globals.HILL_HEIGHT))
					{
						PlaceBarbarianCamp(tile);
						extraCampCount++;
					}
					else if (0 == random.Next(8) && (CountAdjacent(tile, IsLand) > 1) && CountAdjacent(tile, IsWater) > 1)
					{
						tile.meHeight = infos.Globals.DEFAULT_HEIGHT;
						tile.meTerrain = TROPICAL_TERRAIN;
						PlaceBarbarianCamp(tile);
						extraCampCount++;
					}
				}
			}
		}
		for (int y = MapHeight / 4 * 3 + extraCampCount; y < MapHeight - 2; y += 2)
		{ //Northern Barbs
			for (int i = 3; i < MapWidth - 3; i += extraCampCount)
			{
				TileData tile = GetTile(i, y);
				if (0 == random.Next(6) && tile.isUrban(infos) && tile.meImprovement != infos.Globals.RUINS_IMPROVEMENT)
				{
					PlaceBarbarianCamp(tile);
					extraCampCount++;
				}
			}
		}
	}

	private void PlaceBarbarianCamp(TileData tile)
	{
		tile.meCitySite = CitySiteType.ACTIVE_RESERVED;
		tile.meTribeSite = infos.Globals.BARBARIANS_TRIBE;
	}

	protected override void PlaceTribesUsingRandomlySelectedMethod()
	{
		int iDiceRoll = random.Next(7);

		switch (iDiceRoll)
		{
			case 0:
				PlaceSixTribesUsingScatterMethod();
				break;
			case 1:
			case 2:
			case 3:
				PlaceFourTribesUsingQuadrantMethod(); //this one works really well given the geography
				break;
			case 4:
				PlaceThreeTribesUsingLatitudeMethod();
				break;
			case 5:
				PlaceThreeTribesUsingLongitudeBlendMethod();
				break;
			case 6:
				PlaceFiveTribesUsingDominanceMethod(8);
				break;
		}
	}
	protected override void SetBoundaryTiles()
	{
		SmoothTerrain();
		// remove unreachable areas--bugfix for the base version where impassible tiles aren't used to define unreachability
		List<List<int>> boundaryAreas = new List<List<int>>();
		BuildAreas(boundaryAreas, null, x => !x.mbBoundary && !x.isImpassable(infos));

		for (int i = 1; i < boundaryAreas.Count; i++)
		{
			foreach (int tileID in boundaryAreas[i])
			{
				TileData t = GetTile(tileID);
				t.mbBoundary = true;
				if (t.meHeight < infos.Globals.VOLCANO_HEIGHT)
				{
					t.meHeight = infos.Globals.MOUNTAIN_HEIGHT;
					t.meTerrain = TUNDRA_TERRAIN;
					t.meVegetation = VegetationType.NONE;
					t.mbRoad = false;
				}
			}
		}

		//final clean up #1. must not mess with mountain vs land vs water
		foreach (var tile in Tiles)
		{
			//lakes bordering coasts are coasts
			if (tile.meHeight == infos.Globals.LAKE_HEIGHT && CountAdjacent(tile, IsSaltWater) > 0)//could be made more efficient
			{
				ConvertArea(tile, infos.Globals.COAST_HEIGHT);
			}
		}


		//final clean up #2: no changes to terrain or height
		foreach (var tile in Tiles)
		{
			if (CountRiverAndWaterEdges(tile) > 5)
			{
				if (CountAdjacent(tile, IsWater) > 4)
					RemoveRiverFromTile(tile);

			}
			if (tile.meTerrain == WATER_TERRAIN)
			{
				RemoveRiverFromTile(tile);
			}
		}
		base.SetBoundaryTiles();
	}
	protected override bool IsValidPlayerStart(TileData tile, PlayerType player, bool bDoMinDistanceCheck = false)
	{
		int x = TileX(tile);
		int y = TileY(tile);
		if (y < MapHeight / 3 && Math.Abs(x - MapWidth / 2) > 10)
			return false;

		else return base.IsValidPlayerStart(tile, player, bDoMinDistanceCheck);
	}
	protected override void AddUnaffiliatedResources()
	{
		int resourceDensityMultiplier = 100;
		switch (ResourceDensity)
		{
			case ResourceDensityType.POOR:
				resourceDensityMultiplier = 67;
				break;
			case ResourceDensityType.RICH:
				resourceDensityMultiplier = 150;
				break;
		}

		using (new UnityProfileScope("DefaultMapScript.AddUnaffiliatedResources"))
		{
			List<TileData> shuffledTiles = new List<TileData>();

			for (int y = 0; y < MapHeight; ++y)
			{
				for (int x = 0; x < MapWidth; ++x)
				{
					TileData tile = GetTile(x, y);

					if (GetClosestCityDistance(tile) > CitySiteMaxResourceDistance)
					{
						shuffledTiles.Add(tile);
					}
				}
			}

			shuffledTiles.Shuffle(random.NextSeed());

			List<ResourceType> shuffledResources = new List<ResourceType>();

			// place at least one instance of each luxury resource
			for (ResourceType eLoopResource = 0; eLoopResource < infos.resourcesNum(); eLoopResource++)
			{
				if (infos.resource(eLoopResource).miProbThousand > 0 && resourceCounts.GetOrDefault(eLoopResource, 0) == 0)
				{
					if (IsResourceCategory(eLoopResource, ResourceCategory.LUXURY))
					{
						shuffledResources.Add(eLoopResource);
					}
				}
			}
			shuffledResources.Shuffle(random.NextSeed());
			foreach (ResourceType eLoopResource in shuffledResources)
			{
				foreach (TileData tile in shuffledTiles)
				{
					if (CanPlaceResource(tile, eLoopResource))
					{
						PlaceResource(tile, eLoopResource);
						break;
					}
				}
			}
			shuffledResources.Clear();

			for (ResourceType eLoopResource = 0; eLoopResource < infos.resourcesNum(); eLoopResource++)
			{
				shuffledResources.Add(eLoopResource);
			}
			float multiplier = 1;
			foreach (TileData tile in shuffledTiles)
			{
				shuffledResources.Shuffle(random.NextSeed());
				if (TileX(tile) + TileY(tile) < MapHeight / 2) //rich mountain range
					multiplier = 5;
				else if (tile.miLatitude / 20 == 2)
				{
					multiplier = 1.2f;
				}
				else
					multiplier = 0.5f;
				foreach (ResourceType eLoopResource in shuffledResources)
				{
					if ((resourceCounts.GetOrDefault(eLoopResource) < (infos.resource(eLoopResource).miMinPerPlayer * players.Count)) ||
						(random.Next(100000) < infos.resource(eLoopResource).miProbThousand * resourceDensityMultiplier * multiplier))
					{
						if (CanPlaceResource(tile, eLoopResource))
						{
							PlaceResource(tile, eLoopResource);
							break;
						}
					}
				}
			}
		}
	}
	protected override int GetCitySiteValue(TileData siteTile, bool bFreshWater = true, bool bCoastal = true, bool bFood = true, bool bIron = true, bool bStone = true, bool bWood = true)
	{
		using (new UnityProfileScope("DefaultMapScript.GetCitySiteValue"))
		{
			CitySiteValue zValue;
			if (!citySiteValues.TryGetValue(siteTile.ID, out zValue))
			{
				zValue.iBaseValue = -1;
				zValue.iFreshWaterScore = 0;
				zValue.iCoastalScore = 0;
				zValue.iFoodScore = 0;
				zValue.iIronScore = 0;
				zValue.iStoneScore = 0;
				zValue.iWoodScore = 0;

				if (IsValidCitySite(siteTile))
				{
					// Tile has passed fail checks. Process the adjT.
					// Optional flags marked as false will reduce score contributions from that asset type to one fourth of normal.

					if (siteTile.meHeight == infos.Globals.HILL_HEIGHT) // Boost score for defensive bonus if city center is on Hills
					{
						zValue.iBaseValue += 500;
					}

					// Fresh water
					if (IsRiver(siteTile))
					{
						zValue.iFreshWaterScore += 1000;
					}
					else if (IsFreshWaterAccess(siteTile))
					{
						zValue.iFreshWaterScore += 500;
					}

					// Coastal
					int iNumAdjacentOcean = CountAdjacent(siteTile, IsSaltWater);
					int iNumAdjacentLakes = CountAdjacent(siteTile, IsLake);
					int iNumAdjacentWater = iNumAdjacentOcean + iNumAdjacentLakes;

					if (iNumAdjacentWater > 3) // Too much adjacent water. Penalize.
					{
						zValue.iBaseValue -= 200 * iNumAdjacentWater; // Direct penalty to overall score. Not affected by bCoastal.
					}
					else if (iNumAdjacentWater > 0) // Good amount of adjacent water. Add score. On coast of ocean is worth the most.
					{
						if (iNumAdjacentWater == 3)
						{
							zValue.iCoastalScore += 150;
						}
						else if (iNumAdjacentOcean > 0)
						{
							zValue.iCoastalScore += 500;
						}
						else
						{
							zValue.iCoastalScore += 100 * (4 - iNumAdjacentWater);
						}
					}

					// Track relevant map variables out to four rings around the candidate site. (Three for Farming value)
					// Value of each score category diminishes with number of that type in range of the city. Values also tend to diminish for being farther away from city center.
					int iHillsCount = 0;
					int iForestCount = 0;
					int iFertileCount = 0;
					int iArableCount = 0;
					int iMeagerCount = 0;
					int iBadCount = 0;
					int iRiverCount = 0;
					int iFreshWaterCount = 0;

					bool bHasHighValueQuarryTileWithinFirstTwoRings = false;
					bool bHasHighValueQuarryTileInThirdRing = false;

					const int maxRadius = 4;
					const int innerRadius = maxRadius - 1;
					List<int> tileRing = new List<int>();
					for (int radius = 1; radius <= maxRadius; ++radius)
					{
						tileRing.Clear();
						infos.utils().calculateTileRingIds(siteTile.ID, radius, MapWidth, MapHeight, tileRing);

						foreach (int loopTileID in tileRing)
						{
							TileData loopTile = GetTile(loopTileID);

							if (!loopTile.mbBoundary && !loopTile.isImpassable(infos))
							{
								// Check soil quality. Only counting farming tiles in first three rings.
								if (loopTile.meTerrain == LUSH_TERRAIN)
								{
									if (radius <= innerRadius)
									{
										iFertileCount++;
										zValue.iFoodScore += (60 / (iFertileCount)) * (6 - radius) / 5;
									}
								}
								else if (loopTile.meTerrain == TEMPERATE_TERRAIN)
								{
									if (radius <= innerRadius)
									{
										iArableCount++;
										zValue.iFoodScore += (40 / (iArableCount)) * (6 - radius) / 5;
									}
								}
								else if (loopTile.meTerrain == TROPICAL_TERRAIN)
								{
									if (radius <= innerRadius)
									{
										iMeagerCount++;
										zValue.iFoodScore += (10 / (iMeagerCount)) * (6 - radius) / 5;
									}
								}
								else if (loopTile.meTerrain == WATER_TERRAIN)
								{
								}
								else // Bad adjT!
								{
									if (radius < innerRadius) // Only penalizing score for bad terrain tiles within first two rings.
									{
										iBadCount++;
										zValue.iBaseValue -= (150 / (iBadCount)) / radius;
									}
								}

								// Check for Hills
								if (loopTile.meHeight == infos.Globals.HILL_HEIGHT && loopTile.meTerrain != DESERT_TERRAIN)
								{
									iHillsCount++;
									zValue.iIronScore += (50 / Math.Max(1, (iHillsCount - 1 / 2))) * (6 - radius) / 5;
								}
								// Check for Forest
								if (loopTile.meVegetation == TREE_2_VEGETATION)
								{
									iForestCount++;
									if (IsRiver(loopTile))
									{
										zValue.iWoodScore += 60 * (6 - radius) / 5;
									}
									else
									{
										zValue.iWoodScore += (50 / Math.Max(1, (iForestCount - 4 / 2))) * (6 - radius) / 5;
									}
								}
								else if (loopTile.meVegetation == TREE_1_VEGETATION)
								{
									zValue.iWoodScore += 5;
								}

								// Check for River
								if (IsRiver(loopTile))
								{
									iRiverCount++;
									zValue.iBaseValue += (50 / (iRiverCount)) * (6 - radius) / 5;
								}
								else if (IsFreshWaterAccess(loopTile)) // Check for Fresh Water that is not from a River.
								{
									iFreshWaterCount++;
									zValue.iFoodScore += 5;
								}

								// Check for Adjacent to Mountain
								int iNumAdjMountains = GetNumAdjacentMountains(loopTile);
								if (iNumAdjMountains > 0 && iNumAdjMountains < (int)DirectionType.NUM_TYPES)
								{
									zValue.iStoneScore += (200 * iNumAdjMountains) * (6 - radius) / 5;
									if (radius < innerRadius)
									{
										bHasHighValueQuarryTileWithinFirstTwoRings = true;
									}
									else
									{
										bHasHighValueQuarryTileInThirdRing = true;
									}
								}

								//check for great rivers
								if (greatRiverTiles.Contains(loopTile))
								{
									zValue.iBaseValue -= 50 / radius;
									if (!loopTile.isWater(infos))
										zValue.iBaseValue -= 600; //we especially don't want to be near land/crossings 
								}
							}
							else
							{
								zValue.iBaseValue -= 20;
							}
						}
					}

					// Boost overall site value if at least one adjT is a boosted quarry adjT. Mountains tend to come in chains, and we want some cities to be near them.
					if (bHasHighValueQuarryTileWithinFirstTwoRings)
					{
						zValue.iStoneScore += 400;
					}
					else if (bHasHighValueQuarryTileInThirdRing)
					{
						zValue.iStoneScore += 200;
					}
				}

				citySiteValues[siteTile.ID] = zValue;
			}
			return zValue.TotalValue(bFreshWater, bCoastal, bFood, bIron, bStone, bWood);
		}
	}
}





